r"""
Gatekeeper authentication.

The Feide API Gatekeeper will provide a set of credentials using :rfc:`7617`
Basic access authentication. The APIs behind Feide will have to check these
credentials to prevent access from *other* Gatekeepers or clients.

A flask middleware extension, :py:class:`BasicAuth`, provides detection,
authentication and error handling for requests. The validation of credentials
is delegated to a callback function:

::

    def check_credentials(username, password):
        return (username, password) == ('AzureDiamond', 'hunter2')

    app = Flask('example')
    basic_auth = BasicAuth(check_credentials, app)

:py:class:`ConfigBackend` is a callable flask extension for use with
:py:class:`BasicAuth` as the validation callback. The callback compares
credentials with usernames and hashed passwords in the flask app configuration:

::

    app = Flask('example')
    creds = ConfigBackend(app)
    basic_auth = BasicAuth(creds, app)

Individual routes can be protected by authentication by decorating them:

::

    @basic_auth.require
    @app.route('/protected')
    def get_protected():
        return f"Hello {basic_auth.username}\n"

Configuration
-------------
``FEIDE_BASIC_REQUIRE`` (bool, default=False)
    If set to True, apps initialized by :py:class:`BasicAuth` will require
    valid Basic auth credentials for all routes.

``FEIDE_BASIC_REALM`` (string, default=None)
    Realm for 401 challenge created by :py:class:`BasicAuth`.

``FEIDE_BASIC_USERS`` (list, default=[])
    A list of valid credentials for use with :py:class:`ConfigBackend`.

    ::

        [('AzureDiamond', 'hunter2'),
         ('user', '$5$rounds=535000$H5x03ObB8WXUUOh…'),]

"""
import functools
import logging

from flask import current_app, request
from werkzeug.exceptions import Forbidden
from werkzeug.exceptions import Unauthorized as _Unauthorized

from . import creds
from . import utils

logger = logging.getLogger(__name__)


class Unauthorized(_Unauthorized):
    """
    401 HTTPException with support for adding a WWW-Authenticate challenge.

    >>> e = Unauthorized(realm="foo")
    >>> e.get_response().headers['WWW-Authenticate']
    'Basic realm="foo"'
    """

    def __init__(self, realm=None, **kwargs):
        super(Unauthorized, self).__init__(**kwargs)
        self.realm = realm

    def get_response(self, *args, **kwargs):
        response = super(Unauthorized, self).get_response(*args, **kwargs)
        if self.realm:
            response.www_authenticate.set_basic(realm=self.realm)
        else:
            response.www_authenticate.set_basic()
        return response


class ConfigBackend(object):
    """
    Basic auth backend for credentials stored in the flask app config.
    """
    CONFIG_USERS_KEY = 'FEIDE_BASIC_USERS'
    CONFIG_USERS_DEFAULT = tuple()

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        app.config.setdefault(self.CONFIG_USERS_KEY, self.CONFIG_USERS_DEFAULT)

    def __call__(self, username, password):
        """ Check username and password against config (hash). """
        for conf_user, conf_crypt in current_app.config[self.CONFIG_USERS_KEY]:
            if not all((conf_user, conf_crypt)):
                continue
            if username == conf_user and creds.verify(password, conf_crypt):
                return True
        return False


class BasicAuth(object):
    """
    Flask extension that provides validation of Basic auth credentials.
    """

    CONFIG_REALM_KEY = 'FEIDE_BASIC_REALM'
    CONFIG_REALM_DEFAULT = None

    CONFIG_REQUIRE_KEY = 'FEIDE_BASIC_REQUIRE'
    CONFIG_REQUIRE_DEFAULT = False

    is_authenticated = utils.ContextAttribute('is_authenticated')

    def __init__(self, callback, app=None):
        self._callback = callback
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        app.config.setdefault(self.CONFIG_REQUIRE_KEY,
                              self.CONFIG_REQUIRE_DEFAULT)
        app.config.setdefault(self.CONFIG_REALM_KEY,
                              self.CONFIG_REALM_DEFAULT)

        @app.before_request
        def check_basic_auth():
            if request.authorization:
                # Request contains auth data, so let's validate
                self.authenticate()
            else:
                self.is_authenticated = False

        @app.before_request
        def require_basic_auth():
            if (request.url_rule
                    and current_app.config[self.CONFIG_REQUIRE_KEY]
                    and not self.is_authenticated):
                self._raise_missing_auth()

    @property
    def username(self):
        if self.is_authenticated:
            return request.authorization.username
        else:
            return None

    def _raise_missing_auth(self):
        realm = current_app.config.get(self.CONFIG_REALM_KEY, None)
        raise Unauthorized(realm=realm)

    def authenticate(self):
        if request.method == 'OPTIONS':
            # Credentials are not required
            return
        if not (request.authorization and
                request.authorization.type == 'basic'):
            # Missing 'Authorization: Basic ...' header
            self.is_authenticated = False
            logger.debug("BasicAuth missing")
            self._raise_missing_auth()
        elif self._callback(request.authorization.username,
                            request.authorization.password):
            # Correct credentials
            self.is_authenticated = True
            logger.info("BasicAuth succeeded for user=%r",
                        request.authorization.username)
        else:
            # Incorrect credentials
            self.is_authenticated = False
            logger.warning("BasicAuth failed for user=%r",
                           request.authorization.username)
            raise Forbidden()

    def check_is_authenticated(self):
        if not self.is_authenticated:
            self._raise_missing_auth()

    def require(self, func):
        """
        A decorator that can be used to protect specific views with HTTP
        basic auth.
        """
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self.check_is_authenticated()
            return func(*args, **kwargs)
        return wrapper
