"""
Collection of helpers that doesn't really fit anywhere else.
"""
import contextlib
import logging

import requests.exceptions
import werkzeug.exceptions
from flask import _request_ctx_stack, request

logger = logging.getLogger(__name__)


class ContextAttribute:
    """
    Use the request context to store object attribute values.
    """

    def __init__(self, name=None):
        self.name = str(name)

    def __repr__(self):
        return '{cls.__name__}(name={obj.name!r})'.format(cls=type(self),
                                                          obj=self)

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        attr = self.mangled_attr(obj)
        return getattr(self.current_context, attr, None)

    def __set__(self, obj, value):
        attr = self.mangled_attr(obj)
        return setattr(self.current_context, attr, value)

    def __delete__(self, obj):
        attr = self.mangled_attr(obj)
        return delattr(self.current_context, attr)

    @property
    def current_context(self):
        return _request_ctx_stack.top

    def mangled_attr(self, owner):
        return '__{cls}_{sep}{addr}'.format(
            cls=type(self).__name__,
            sep='{}_'.format(self.name) if self.name else '',
            addr=id(owner))


@contextlib.contextmanager
def upstream_request_handler(service="a third party service"):
    """
    Context to map upstream http error codes.

    This context will catch `requests.RequestException` and re-raise as an
    appropriate `werkzeug.exceptions.HTTPException` for the client.

    Typical usage:

    ::

        @app.route('/foo')
        def get_foo():
            with upstream_request_handler():
                return requests.get('http://example.org/foo').json()
    """
    try:
        yield
    except requests.exceptions.Timeout:
        logger.error("timeout in request to %s", service, exc_info=True)
        raise werkzeug.exceptions.GatewayTimeout("Unable to contact %s" %
                                                 (service, ))
    except requests.exceptions.RequestException:
        logger.error("error in request to %s", service, exc_info=True)
        raise werkzeug.exceptions.BadGateway("Unable to contact %s" %
                                             (service, ))


def get_multi_header(name, sep=',', strip=True):
    """ Get a multi-value, comma separated header value.

    :rfc:`2616` states that certain headers may be duplicated, and in those
    cases that the header values *must* be comma-separated lists that could be
    joined together in one list.

    Using a wsgi server, this join must be performed before we ever see the
    request data - but in other environments, the
    :py:data:`flask.request.headers` object *could* contain multiple values for
    a given header name.
    """
    return tuple(
        (item.strip() if strip else item)
        for value in request.headers.getlist(name)
        for item in value.split(sep))


def format_block(strblock, prefix="", header="", indent=0):
    """ indent and prefix lines of text. """
    def fmt_line(line):
        return '{!s}{!s}{!s}'.format(' ' * indent, prefix, line)

    def fmt_header(line):
        return "{!s}{!s}\n".format(' ' * indent, line) if line else ''

    return "{!s}{!s}\n".format(
        fmt_header(header),
        "\n".join(fmt_line(line) for line in strblock.split("\n")))


def request_to_str(request, omit_body=False):
    """ Format a requests.Request object as a raw request string. """
    parts = [
        # May not be entirely correct, as we don't have access to the actual
        # session data or connection adapter.
        "{r.method!s} {r.url} HTTP/1.x".format(r=request),
        "\n".join(("{!s}: {!s}".format(h, v))
                  for h, v in request.headers.items()),
        "",
    ]
    if request.body:
        if omit_body:
            parts.append("[+ {:d} chars]".format(len(request.body)))
        else:
            parts.append(request.body)
    return "\n".join(parts)


def response_to_str(response, omit_body=False):
    """ Format a requests.Response object as a raw response string. """
    parts = [
        "{r.status_code:d} {r.reason!s}".format(r=response),
        "\n".join(("{!s}: {!s}".format(h, v))
                  for h, v in response.headers.items()),
        "",
    ]
    if response.content:
        if omit_body:
            parts.append("[+ {:d} bytes]".format(len(response.content)))
        else:
            parts.append(response.text)
    return "\n".join(parts)
